from __future__ import print_function, division, absolute_import, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

import numpy
import os.path

from .bdf import BDF, ant2bl, bl2ant


# higher-level class to gather per-scan info from SDM and BDF sets

def uid2fname(s):
    """Convert uid URL to file name (mainly for BDFs)."""
#    return s.translate(string.maketrans(':/', '__'))
    return s.replace(':/', '__').replace('/', '_')


def sdmarray(s, dtype=None):
    """Convert an array-valued SDM entry (string) into a numpy array."""
    fields = str(s).split()
    ndim = int(fields[0])
    dims = tuple(map(int, fields[1:ndim+1]))
    return numpy.array(fields[ndim+1:], dtype=dtype).reshape(dims)


class Scan(object):
    """
    Represents a single subscan as part of a SDM/BDF dataset.  Convenience
    interface to open the BDF, get useful metadata, etc.
    """
    def __init__(self, sdm, scanidx, subscanidx=1):
        """
        sdm is the SDM object.
        scanidx is the scan number.
        subscanidx is the subscan number.
        """
        self.sdm = sdm
        self.idx = str(scanidx)
        self.subidx = str(subscanidx)
        self._bdf = None
        self.__main = None
        self.__scan = None
        self.__subscan = None

    @property
    def bdf(self):
        if self._bdf is None:
            self._bdf = BDF(self.bdf_fname)
        return self._bdf

    @property
    def _main(self):
        """Convenience interface to the SDM Main table row."""
        if self.__main is None:
            self.__main = self.sdm['Main'][(self.idx, self.subidx)]
        return self.__main

    @property
    def _scan(self):
        """Convenience interface to the SDM Scan table row."""
        if self.__scan is None:
            self.__scan = self.sdm['Scan'][self.idx]
        return self.__scan

    @property
    def _subscan(self):
        """Convenience interface to the SDM Subscan table row."""
        if self.__subscan is None:
            self.__subscan = self.sdm['Subscan'][(self.idx,self.subidx)]
        return self.__subscan

    @property
    def _config(self):
        """Convenience interfact to the SDM ConfigDescription row."""
        return self.sdm['ConfigDescription'][self._main.configDescriptionId]

    @property
    def bdfdir(self):
        return self.sdm.bdfdir if self.sdm.bdfdir \
            else os.path.join(self.sdm.path, 'ASDMBinary')

    @property
    def bdf_fname(self):
        try:
            bdf_fname = os.path.join(self.bdfdir,
                                     uid2fname(self._main.dataUID.EntityRef.get('entityId')))
        except AttributeError:
            bdf_fname = os.path.join(self.bdfdir,
                                     uid2fname(self._main.dataOid.EntityRef.get('entityId')))
        return bdf_fname

    @property
    def source(self):
        """Source name as defined in SDM Scan table."""
        return str(self._scan.sourceName)

    @property
    def field(self):
        """Field name as defined in SDM Subscan table."""
        return str(self._subscan.fieldName)

    @property
    def coordinates(self):
        """
        Return the pointing coordinates (radians as given in Field table).
        """

        # Note as usual there are many redundant choices for where to
        # get this info from the SDM.  This is probably fine for standard
        # VLA observations where each scan points at a single location.
        # It may not be what is desired for OTF (mapping) type observations.
        # The SDM also seems to support some kind of polynomial in the Field
        # table; here we just return the 0th order part of this.

        return sdmarray(self.sdm['Field'][self._main.fieldId].referenceDir,
                        dtype=numpy.float)[0]

    @property
    def intents(self):
        """Return the list of intents for this scan."""
        return list(sdmarray(self._scan.scanIntent))

    @property
    def subintent(self):
        """Return the subscan intent."""
        return str(self._subscan.subscanIntent)

    @property
    def antennas(self):
        """Return the list of antenna names for this scan."""
        sdm_ants = sdmarray(self._config.antennaId)
        return [str(self.sdm['Antenna'][a].name) for a in sdm_ants]

    @property
    def stations(self):
        """Return the list of station names for this scan."""
        sdm_ants = sdmarray(self._config.antennaId)
        sdm_stns = [self.sdm['Antenna'][a].stationId for a in sdm_ants]
        return [str(self.sdm['Station'][s].name) for s in sdm_stns]

    @property
    def positions(self):
        """
        Return the list of antenna posisitons (XYZ, m) for this scan.
        Result is an nant-by-3 array.
        """

        sdm_ants = sdmarray(self._config.antennaId)
        sdm_stns = [self.sdm['Antenna'][a].stationId for a in sdm_ants]
        return [sdmarray(self.sdm['Station'][s].position, dtype=numpy.float)
                for s in sdm_stns]

    @property
    def baselines(self):
        """
        Return the list of antenna pairs for this scan, in BDF ordering.
        """

        ants = self.antennas
        nant = len(ants)
        nbl = nant*(nant-1)//2
        # return ['%s-%s' % (ants[x[0]], ants[x[1]])
        #        for x in map(bl2ant, range(nbl))]
        return [(ants[x[0]], ants[x[1]])
                for x in map(bl2ant, list(range(nbl)))]

    @property
    def startMJD(self):
        return float(self._subscan.startTime/86400.0e9)

    @property
    def endMJD(self):
        return float(self._subscan.endTime/86400.0e9)

    @property
    def numIntegration(self):
        """Number of integrations as listed in the SDM Main table."""
        return int(self._main.numIntegration)

    @property
    def spws(self):
        """ Return the list of spw names """

        return [str(self.sdm['DataDescription'][dd_id].spectralWindowId)
                for dd_id in sdmarray(self._config.dataDescriptionId)]

    @property
    def reffreqs(self):
        """ List of reference frequencies. One per spw in spws list. """

        return [float(self.spw(spwn).refFreq)
                for spwn in range(len(self.spws))]

    @property
    def numchans(self):
        """ List of number of channels per spw. One per spw in spws list. """

        return [int(self.spw(spwn).numChan) for spwn in range(len(self.spws))]

    @property
    def chanwidths(self):
        """ List of channel widths. One per spw in spws list. """

        return [float(self.spw(spwn).chanWidth)
                for spwn in range(len(self.spws))]

    @property
    def pulsar(self):
        """ Pulsar row entry, if one exists, otherwise None. """
        # Note assumes spw 0 right now (in principle could be different
        # pulsar models per spw).
        try:
            dd_id = sdmarray(self._config.dataDescriptionId)[0]
            psr_id = self.sdm['DataDescription'][dd_id].pulsarId
            return self.sdm['Pulsar'][psr_id]
        except AttributeError:
            return None

    def times(self, src='bdf'):
        """Return array of midpoint times (MJD) for all integrations in this 
        scan.  If src=='sdm' these will be estimated from the information in
        the SDM Subscan table.  If src=='bdf' they will be read from the BDF.
        If precise times are needed, use the BDF values, but this requires
        reading the BDF which can be slow."""
        if src.lower()=='sdm':
            t0 = self.startMJD
            ni = int(self._subscan.numIntegration)
            dt = (self.endMJD - self.startMJD) / ni
            return t0 + (numpy.arange(ni)+0.5)*dt
        elif src.lower()=='bdf':
            return numpy.array([i.time for i in self.bdf])
        else:
            raise ValueError("Unknown data source: " + str(src))

    def freqs(self, spwidx='all'):
        """ Array of per-channel frequences for the given spectral window.
        If spwidx=='all', a nspw-by-nchan array will be returned giving all
        frequencies, if all spectral window have the same number of channels.
        """

        nspw = len(self.spws)
        rf = self.reffreqs
        nc = self.numchans
        cw = self.chanwidths
        if spwidx == 'all':
            if nc.count(nc[0]) != len(nc):
                raise RuntimeError("Variable number of channels")
            nc = nc[0]
            out = numpy.zeros((nspw, nc))
            for i in range(nspw):
                out[i, :] = numpy.arange(nc)*cw[i] + rf[i]
        else:
            out = numpy.arange(nc[spwidx]) * cw[spwidx] + rf[spwidx]
        return out

    def tcal(self,spwidx='all'):
        """ Array of Tcal values from CalDevice table."""
        sdm_ants = sdmarray(self._config.antennaId)
        nspw = len(self.spws)
        nant = len(sdm_ants)
        if spwidx=='all':
            out = numpy.zeros((nant,nspw,2)) # assumes 2-pol values..
            for iant in range(nant):
                for ispw in range(nspw):
                    tmp = self.sdm['CalDevice'][(sdm_ants[iant],
                        self.spws[ispw])].coupledNoiseCal
                    out[iant,ispw,:] = sdmarray(tmp,dtype=numpy.float)[:,0]
        else:
            raise NotImplementedError('Single spw not implemented yet')
        return out

    def spw(self, idx):
        """Return the SpectralWindow entry for the given index in this scan."""

        dd_id = sdmarray(self._config.dataDescriptionId)[idx]
        spw_id = self.sdm['DataDescription'][dd_id].spectralWindowId
        return self.sdm['SpectralWindow'][spw_id]

    def flags(self, mjd, axis='bl', pad=False, flagval=0, expand=1.0):
        """
        Return flag array for the given time(s).  Input mjd can be scalar
        or array valued.  If axis=='ant', returned array will have dimensions
        (N_times, N_antenna), otherwise (N_times, N_baselines).  Flag array
        contains 1 for non-flagged points and flagval for flagged data.  If pad
        arg is true, four extra len-1 dimensions will be appended so that
        flags can be applied to standard bdf.get_data() results.  All
        flag start/stop times are increased by the value of the expand
        argument (seconds).
        """

        sdm_ants = self.antennas
        nant = len(sdm_ants)
        nbl = nant*(nant-1)//2
        t_ns = numpy.array(numpy.array(mjd)*86400.0e9, dtype=numpy.int64)
        exp_ns = int(expand*1e9)
        d_out = t_ns.shape
        isscalar = (d_out == ())
        if axis == 'ant':
            d_out += (nant,)
        else:
            d_out += (nbl,)
        if pad:
            d_out += (1, 1, 1, 1)
        out = numpy.ones(d_out, dtype=type(flagval))
        for flag in self.sdm['Flag']:
            tidx = numpy.where((t_ns > (int(flag.startTime)-exp_ns)) *
                               (t_ns < (int(flag.endTime)+exp_ns)))[0]
            if len(tidx) == 0:
                continue
            for a in sdmarray(flag.antennaId):
                flagant = self.sdm['Antenna'][a].name
                if flagant not in sdm_ants:
                    continue
                if axis == 'ant':
                    flagidx = [sdm_ants.index(flagant), ]
                else:
                    flagidx = numpy.where([flagant in pair
                                          for pair in self.baselines])[0]
                if isscalar:
                    out[flagidx] = flagval
                else:
                    for ii in flagidx:
                        out[tidx, ii] = flagval
        return out



#! /usr/bin/env python

# Average SDM/BDF data in time
# PBD 05/2016

import os, sys
import numpy as np
np.errstate(divide='ignore')
import sdmpy
import progressbar

import argparse
par = argparse.ArgumentParser()
par.add_argument("sdmname", help="SDM to process")
par.add_argument("-t", "--time", type=float, default=3.0,
        help="averaging time (s) [%(default)s]")
par.add_argument("-e", "--ext", default="avg",
        help="extension to add to output SDM [%(default)s]")
par.add_argument("-b", "--bdfdir", default="",
        help="path to BDFs (optional)")
par.add_argument("-f", "--flagfrac", default=1.0, type=float,
        help="Flag integrations with zero fraction greater than flagfrac")
args = par.parse_args()

sdmname = args.sdmname.rstrip('/')

sdm = sdmpy.SDM(sdmname,bdfdir=args.bdfdir)

assert (args.flagfrac >= 0.) and (args.flagfrac <= 1.)

tavg = args.time # in seconds

sdmout = sdmname + '.' + args.ext
bdfoutpath = sdmout + '/ASDMBinary'

os.mkdir(sdmout)
os.mkdir(bdfoutpath)

for scan in sdm.scans():
    print("Processing '{0}' scan {1}:".format(sdmname, scan.idx))
    if scan.bdf.fp is None:
        print("Error reading bdf for scan {0}, skipping".format(scan.idx))
        continue

    bdf = scan.bdf

    # TODO maybe rename BDFs...
    # Also this assumes averaging time is the same during the whole
    # BDF.  This is always true for VLA data.
    bdfoutname = bdfoutpath + '/' + os.path.basename(scan.bdf_fname)
    navg = max(1, int(tavg / bdf[0].interval))
    nout = int(bdf.numIntegration / navg)
    delta_t = int((navg/2.0)*bdf[0].interval*1e9) # ns
    # Set up for output BDF, copying header info from the input BDF
    bdfout = sdmpy.bdf.BDFWriter(bdfoutpath, fname=os.path.basename(scan.bdf_fname), bdf=bdf)
    bdfout.write_header()
    bar = progressbar.ProgressBar()
    zeroed = 0
    for i in bar(range(nout)):
        bdfint = bdf[i*navg]
        count = {}
        for dtype in bdfint.data.keys():
            bdfint.data[dtype] = bdfint.data[dtype].copy()
            count[dtype] = np.zeros(bdfint.data[dtype].shape)
            count[dtype] += bdfint.data[dtype] != 0.0
        zerofrac = 0.
        for j in range(1, navg):
            if args.flagfrac < 1.0:
                zerofrac += bdf[i*navg+j].zerofraction()  # default looks at cross
            for dtype in bdfint.data.keys():
                dat = bdf[i*navg+j].data[dtype]
                bdfint.data[dtype] += dat
                count[dtype] += dat != 0.0
        zerofrac /= navg
        for dtype in bdfint.data.keys():
            bdfint.data[dtype] /= count[dtype]
            bdfint.data[dtype][np.where(count[dtype]==0.0)] = 0.0
            if zerofrac > args.flagfrac:
                bdfint.data[dtype][:] = 0.0  # flag data if zeros exceed limit
                zeroed += 1

        # update timestamp and interval
        bdfint.sdmDataSubsetHeader.schedulePeriodTime.time += delta_t
        bdfint.sdmDataSubsetHeader.schedulePeriodTime.interval *= navg
        bdfout.write_integration(bdfint)
    bdfout.close()

    if zeroed:
        print("Flagged {0} integrations for zerofrac exceeding flagfrac.".format(zeroed))

    # update SDM entries with corect number of integrations, etc.
    scan._main.numIntegration = nout
    scan._main.dataSize = os.path.getsize(bdfoutname)
    scan._subscan.numIntegration = nout
    scan._subscan.numSubintegration = ('1 %d' % nout) + ' 0'*nout

# Write out new SDM tables
sdm.write(sdmout)




#! /usr/bin/env python
import sys, math
import numpy as np
import logging
import sdmpy
import sdmpy.pulsar
import psrchive
#import tempo_utils

# Only needed for changing direction
try:
    import casatools
except ImportError:
    casatools = None

import argparse
par = argparse.ArgumentParser(
        description="Convert phase-binned SDM/BDF data to fold-mode PSRFITS.")
par.add_argument("sdmname", help="SDM to process")
#par.add_argument("-v", "--verbose", dest='loglevel',
#        action='store_const', const=logging.DEBUG,
#        default=logging.INFO,
#        help='Enable DEBUG output')
par.add_argument("-q", "--quiet", dest='loglevel',
        action='store_const', const=logging.WARN,
        default=logging.INFO,
        help='No INFO messages')
par.add_argument("-d", "--dm", type=float, default=0.0,
        help="dispersion measure (pc cm^-3) [%(default)s]")
par.add_argument("-p", "--period", type=float, default=0.0,
        help="period data were folded (s) [auto]")
par.add_argument("-P", "--polycos", action="store_true",
        help="use polyco file to adjust timestamps [false]")
par.add_argument("-H", "--hanning", action="store_true",
        help="apply Hanning smoothing")
par.add_argument("-D", "--direction", type=str, default='',
        help="rephase data to specified ra,dec")
args = par.parse_args()

logging.basicConfig(format="%(asctime)-15s %(levelname)8s %(message)s", 
        level=args.loglevel)

if args.direction:
    if casatools is None:
        logging.error('Rephasing sky direction requires CASA')
        sys.exit(1)
    # Could use astropy/etc to parse coords but we already 
    # require CASA..
    qa = casatools.quanta()
    ra, dec = args.direction.split(',')
    bad_coords = False
    try:
        ra_deg = qa.quantity(ra)
        dec_deg = qa.quantity(dec)
    except RuntimeError:
        bad_coords = True
    if (ra_deg['unit']!='deg' or dec_deg['unit']!='deg'
            or bad_coords):
        logging.error('Error parsing coordinates, try eg 12h34m56.7s,+12d34m56.7s')
        sys.exit(1)
    ra_rad = qa.convert(ra_deg,'rad')['value']
    dec_rad = qa.convert(dec_deg,'rad')['value']

sdmname = args.sdmname.rstrip('/')
sdm = sdmpy.SDM(sdmname, use_xsd=False)
try:
    binlog = sdmpy.pulsar.BinLog(sdmname)
except IOError:
    binlog = None

# TODO update to read 4-pol data
npol = 2

def unroll_chans(data):
    """Unroll the spw and channel dims into a single frequency axis.
    Assumes incoming dimensions are (bl, spw, bin, chan, pol)"""
    data = data.transpose((0,2,1,3,4))
    data = data.reshape((data.shape[0], 
        data.shape[1], 
        data.shape[2]*data.shape[3],
        data.shape[4]))
    return data

for scan in sdm.scans():
    if not scan.bdf.exists: continue
    if 'CALIBRATE_PHASE' in scan.intents:
        logging.info('processing cal scan %s' % scan.idx)
        # dims (bl,spw,bin,chan,pol)
        dcal = scan.bdf.get_data(scrunch=True)[...,[0,-1]].mean(2,
                keepdims=True)
        if args.hanning:
            sdmpy.calib.hanning(dcal,axis=3)
        dcal = unroll_chans(dcal)
        gcal = sdmpy.calib.gaincal(dcal,axis=0,ref=1)
    elif 'OBSERVE_TARGET' in scan.intents:
        logging.info('processing target scan %s' % scan.idx)

        psr = str(scan.source)

        # Assumes all spws have same number of bins
        nbin = int(scan.bdf.spws[0].numBin)

        # Assumes all spws have same number of chans
        freqs = scan.freqs().ravel()/1e6
        chanidx = np.argsort(freqs)
        nchan = freqs.size
        bw = sum([scan.spw(i).totBandwidth/1e6 for i in range(len(scan.spws))])

        # Initialize archive output
        arch = psrchive.Archive_new_Archive("ASP")
        arch.resize(0,npol,nchan,nbin)
        arch.set_source(psr)
        arch.set_dispersion_measure(args.dm)
        arch.set_coordinates(psrchive.sky_coord(*scan.coordinates))
        arch.set_centre_frequency(0.5*(freqs.max() + freqs.min()))
        arch.set_bandwidth(bw)
        arch.set_telescope('vla')
        # This fails in python3 for unknown reasons:
        #arch.set_state('PPQQ')
        # Workaround:
        arch.execute('e state=PPQQ')

        iout = 0

        # Try to load polycos
        try:
            if scan.pulsar is not None:
                polys = sdmpy.pulsar.sdmpulsar_to_polyco(scan.pulsar,
                        fmt='psrchive')
                # Used for testing rephasing:
                #polys0 = sdmpy.pulsar.sdmpulsar_to_polyco(scan.pulsar)
                #polys1 = tempo_utils.polycos.generate_from_polyco(
                #        "/users/pdemores/tzpar/B1937+21.par",
                #        polys0)
                logging.info('Read polycos from SDM Pulsar table')
            else:
                polycofile = '%s/%s.%d.polyco' % (binlog._logdir,
                        sdmname, int(scan.idx))
                polys = psrchive.polyco(polycofile)
                logging.info('Read polycos from %s' % polycofile)
        except Exception as ex:
            logging.info("Couldn't load polycos: " + repr(ex))
            polys = None

        bdf = scan.bdf
        arch.resize(arch.get_nsubint() + bdf.numIntegration - 1)
        for isub in range(1,bdf.numIntegration):
            logging.info("Processing subint %d/%d" % (isub,bdf.numIntegration))
            bdfsub = bdf[isub]
            dpsr = bdfsub.get_data()[...,[0,-1]]

            if args.hanning:
                sdmpy.calib.hanning(dcal,axis=3)

            dpsr = unroll_chans(dpsr)
            sdmpy.calib.applycal(dpsr,gcal,phaseonly=True)

            if args.direction:
                # Rephase data to a new sky direction
                logging.info("Rephasing sky direction")
                uvw0 = sdmpy.calib.uvw(bdfsub.time, scan.coordinates, 
                        scan.positions)
                uvw1 = sdmpy.calib.uvw(bdfsub.time, (ra_rad, dec_rad),
                        scan.positions)
                dw_us = 1e6*(uvw1[:,2] - uvw0[:,2])/299792458.0
                phs = np.outer(dw_us,freqs)
                phs = phs.reshape((phs.shape[0],1,phs.shape[1],1))
                dpsr *= np.exp(2.0j*np.pi*phs) # pos sign correct

            # Sum over baselines:
            dpsr = np.ma.masked_array(dpsr,dpsr==0.0)
            mpsr = np.real(dpsr.mean(0))

            subint = arch.get_Integration(iout)
            epoch_bdf = psrchive.MJD(float(bdfsub.time)) # only approx
            if args.period>0.0:
                #epoch = psrchive.MJD(epoch_bdf.intday())
                epoch = epoch_bdf
                p = args.period
                dt = 0.0
                logging.info('Using constant period')
            else:
                if (binlog is not None) or (polys is not None):
                    # Apply polyco time adjust
                    if args.polycos and polys is not None:
                        p = polys.period(epoch_bdf)
                        dt = (polys.period(epoch_bdf) 
                                * polys.phase(epoch_bdf).fracturns())
                        epoch = epoch_bdf - dt
                    else:
                        (epoch,p,dt) = binlog.epoch_period(epoch_bdf)
                else:
                    (epoch,p,dt) = sdmpy.pulsar._get_epoch_period(epoch_bdf)
                logging.info('Using epoch/period from dt=%.3fs' % dt)

            # These were used for testing dedispersion/rephasing:
            ## This sign is correct:
            #dphase = polys1.phase(bdfsub.time) - polys0.phase(bdfsub.time)
            #logging.warning("Testing rephasing p=%.6f dphase=%.3f"%(p,dphase))
            #sdmpy.pulsar.dedisperse_array(mpsr, args.dm, freqs, p,
            #        bin_axis=0, freq_axis=1, phase_shift=dphase)
            #sdmpy.pulsar.dedisperse_array(mpsr, args.dm, freqs_spw, p,
            #        bin_axis=2,freq_axis=3,spw_axis=1)
            #mpsr = unroll_chans(mpsr)[0,...]

            subint.set_epoch(epoch)
            subint.set_duration(bdfsub.interval)
            if p>0.0: subint.set_folding_period(p)
            if isub==0:
                wt=0.0
            else:
                wt=1.0
            for ochan in range(nchan):
                ichan = chanidx[ochan]
                subint.set_centre_frequency(ochan,freqs[ichan])
                for ipol in range(npol):
                    prof = subint.get_Profile(ipol,ochan)
                    prof.get_amps()[:] = mpsr[:,ichan,ipol]
                subint.set_weight(ochan,wt)
            iout += 1

        if polys is not None:
            arch.set_model(polys,False)
        outputname = sdmname + '.%03d.fits' % int(scan.idx)
        logging.info("unloading '%s'" % outputname)
        arch.unload(outputname)



from __future__ import print_function, division, absolute_import, unicode_literals # not casa compatible
from builtins import bytes, dict, object, range, map, input#, str # not casa compatible
from future.utils import itervalues, viewitems, iteritems, listvalues, listitems
from io import open

import numpy as np
from numpy import linalg

from .bdf import ant2bl, bl2ant

# Some routines to derive simple calibration solutions directly
# from data arrays.

def gaincal(data, axis=0, ref=0, avg=[], nit=3):
    """Derives amplitude/phase calibration factors from the data array
    for the given baseline axis.  In the returned array, the baseline
    dimension is converted to antenna.  No other axes are modified.
    Note this internally makes a transposed copy of the data so be
    careful with memory usage in the case of large data sets.  A list
    of axes to average over before solving can be given in the avg
    argument (length-1 dimensions are kept so that the solution can be
    applied to the original data)."""
    nbl = data.shape[axis]
    ndim = len(data.shape)
    (check, nant) = bl2ant(nbl)
    if check != 0:
        raise RuntimeError("Specified axis dimension (%d) is not a valid number of baselines" % nbl)
    if avg != []:
        # Average, ignoring zeros
        #norm = np.count_nonzero(data,axis=avg,keepdims=True) # requires numpy 1.19 for keepdims
        norm = np.count_nonzero(data,axis=tuple(avg))
        norm[np.where(norm==0)] = 1
        data = data.sum(axis=tuple(avg), keepdims=True)
        norm = norm.reshape(data.shape) # workaround lack of keepdims
        data = data / norm
    tdata = np.zeros(data.shape[:axis]+data.shape[axis+1:]+(nant, nant),
                     dtype=data.dtype)
    for i in range(nbl):
        (a0, a1) = bl2ant(i)
        tdata[..., a0, a1] = data.take(i, axis=axis)
        tdata[..., a1, a0] = np.conj(data.take(i, axis=axis))
    for it in range(nit):
        (wtmp, vtmp) = linalg.eigh(tdata)
        v = vtmp[..., -1].copy()
        w = wtmp[..., -1]
        for i in range(nant):
            tdata[..., i, i] = w*(v.real[..., i]**2 + v.imag[..., i]**2)
    # result = np.sqrt(w[...,-1]).T*v[...,-1].T
    result = np.sqrt(w).T*v.T
    # First axis is now antenna.. refer all phases to reference ant
    result = (result*np.conj(result[ref])/np.abs(result[ref])).T
    # TODO try to reduce number of transposes
    outdims = list(range(axis)) + [-1, ] + list(range(axis, ndim-1))
    return result.transpose(outdims)


def applycal(data, caldata, axis=0, phaseonly=False):
    """
    Apply the complex gain calibration given in the caldata array
    to the data array.  The baseline/antenna axis must be specified in
    the axis argument.  Dimensions of all other axes must match up
    (in the numpy broadcast sense) between the two arrays.
    """

    ndim = len(data.shape)
    nbl = data.shape[axis]
    nant = caldata.shape[axis]
    (check, nant_check) = bl2ant(nbl)
    if check != 0:
        raise RuntimeError("Specified axis dimension (%d) is not a valid number of baselines" % nbl)
    if nant != nant_check:
        raise RuntimeError("Number of antennas does not match (data=%d, caldata=%d)" % (nant_check, nant))
    if phaseonly:
        caldata = caldata.copy()/abs(caldata)
        caldata[np.where(np.isfinite(caldata) == False)] = 0.0j
    icaldata = 1.0/caldata
    icaldata[np.where(np.isfinite(icaldata) == False)] = 0.0j
    # Modifies data in place.  Would it be better to return a calibrated
    # copy instead of touching the original?
    for ibl in range(nbl):
        # Must be some cleaner way to do this..?
        dslice = (slice(None),)*axis + (ibl,) + (slice(None),)*(ndim-axis-1)
        (a1, a2) = bl2ant(ibl)
        calfac =  icaldata.take(a1, axis=axis) * icaldata.take(a2, axis=axis).conj()
        data[dslice] *= calfac

def hanning(data, axis=0):
    """Apply hanning smoothing along the specified axis, typically this should
    be the spectral channel axis.  Modifies data array in-place."""
    data_pos = np.roll(data,1,axis=axis)
    data_neg = np.roll(data,-1,axis=axis)
    data += 0.5*(data_pos + data_neg)
    data *= 0.5

def uvw(mjd, direction, antpos, method="casa"):
    """Return an Nbaseline-x-3 array giving U,V,W in meters for
    the given MJD, sky direction, and antenna positions.

      direction is (ra,dec) in radians
      antpos is Nant-by-3 array of antenna positions in meters.
      method can be "casa" or "astropy"

    This can be called on an sdmpy Scan object like:

      uvw = sdmpy.calib.uvw(scan.startMJD, scan.coordinates, scan.positions)
    """
    if method=="casa":
        return _uvw_casa(mjd, direction, antpos)
    elif method=="astropy":
        return _uvw_astropy(mjd, direction, antpos)
    else:
        raise RuntimeError("uvw method '%s' unknown" % method)

# Calculate uvw using CASA, mostly copied
# from rfpipe's calc_uvw().
def _uvw_casa(mjd, direction, antpos):
    """Return an Nbaseline-x-3 array giving U,V,W in meters for
    the given MJD, sky direction, and antenna positions.

      direction is (ra,dec) in radians
      antpos is Nant-by-3 array of antenna positions in meters.

    This can be called on an sdmpy Scan object like:

      uvw = sdmpy.calib.uvw(scan.startMJD, scan.coordinates, scan.positions)
    """
    import casatools
    me = casatools.measures()
    qa = casatools.quanta()
    qq = qa.quantity
    s = me.direction('J2000',
            qq(direction[0],'rad'),
            qq(direction[1],'rad'))
    e = me.epoch('UTC', qq(mjd,'d'))
    o = me.observatory('VLA')
    me.doframe(o)
    me.doframe(e)
    me.doframe(s)
    pos = np.array(antpos)
    casapos = me.position('ITRF',
            qq(pos[:,0],'m'),
            qq(pos[:,1],'m'),
            qq(pos[:,2],'m'))
    bls = me.expand(me.touvw(me.asbaseline(casapos))[0])[1]['value']
    bls = bls.reshape((-1,3))
    # Original code from rfpipe:
    #ord1 = [(i,j) for i in range(nants) for j in range(i+1, nants)] # CASA order
    #ord2 = [(i,j) for j in range(nants) for i in range(j)]  # BDF order
    nant = pos.shape[0]
    nbl = bls.shape[0]
    # Index into the output (BDF-style) array for each value 
    # in the input (CASA-style) array.
    oidx = [ant2bl((i,j)) for i in range(nant) for j in range(i+1,nant)]
    uvw = 0.0*bls
    for i in range(nbl):
        uvw[oidx[i],:] = bls[i,:]
    return uvw

def _uvw_astropy(mjd, direction, antpos):
    """ Calculates and returns uvw in meters for a given time and pointing direction.
    direction is (ra,dec) as tuple in radians.
    Can optionally specify a telescope other than the VLA.
    """
    from astropy import coordinates, time

    telescope = 'VLA'

    phase_center = coordinates.SkyCoord(*direction, unit='rad', frame='icrs')

    antpos = np.array(antpos)
    antpos = coordinates.EarthLocation(x=antpos[:,0], y=antpos[:,1], z=antpos[:,2], unit='m')

    datetime = time.Time(mjd,format='mjd')

    tel_p, tel_v = coordinates.EarthLocation.of_site(telescope).get_gcrs_posvel(datetime)
    antpos_gcrs = coordinates.GCRS(antpos.get_gcrs_posvel(datetime)[0],
                                   obstime = datetime, obsgeoloc = tel_p,
                                   obsgeovel = tel_v)

    uvw_frame = phase_center.transform_to(antpos_gcrs).skyoffset_frame()
    antpos_uvw = antpos_gcrs.transform_to(uvw_frame).cartesian

    nant = len(antpos_uvw)
    antpairs = [(i,j) for j in range(nant) for i in range(j)]
    nbl = len(antpairs)
    uvw = np.empty((nbl,3))
    for ibl, ant in enumerate(antpairs):
        bl = antpos_uvw[ant[1]] - antpos_uvw[ant[0]]
        uvw[ibl,0] = bl.y.value
        uvw[ibl,1] = bl.z.value
        uvw[ibl,2] = bl.x.value

    return uvw

def rephased_data(scan, radec, uvw_method="casa"):
    """
    Return all data for the scan, rephased to new coordinates.

    radec = new (ra, dec) in radians
    uvw_method = "casa" or "astropy"
    """
    freqs = scan.freqs().ravel() / 1e6
    dat = None
    for isub in range(scan.bdf.numIntegration):
        bdfsub = scan.bdf[isub]
        subdat = bdfsub.get_data().copy()
        uvw0 = uvw(bdfsub.time, scan.coordinates, scan.positions, uvw_method)
        uvw1 = uvw(bdfsub.time, radec, scan.positions, uvw_method)
        dw_us = 1e6 * (uvw1[:,2] - uvw0[:,2])/299792458.0
        phs = np.outer(dw_us,freqs).reshape((subdat.shape[0],subdat.shape[1],1,subdat.shape[3],1))
        subdat *= np.exp(2.0j*np.pi*phs)
        if dat is None:
            dat = np.expand_dims(subdat,0)
        else:
            dat = np.concatenate((dat,np.expand_dims(subdat,0)),axis=0)
    return dat



